import os
import sys

import pytest

import unittest.mock as mock

from atpbar.presentation.detect.spy import is_spyder_ide

##__________________________________________________________________||
@pytest.fixture()
def mock_spyder_module(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['atpbar.presentation.detect.spy']
    monkeypatch.setattr(module, 'spyder', ret)
    yield ret

@pytest.fixture()
def mock_get_ipython(monkeypatch):
    mock_ipython = mock.Mock()
    mock_ipython.config = {
        'IPKernelApp': {
            'connection_file': '/some/path/kernel.json',
        },
    }
    ret = mock.Mock()
    ret.return_value = mock_ipython

    module = sys.modules['atpbar.presentation.detect.spy']
    monkeypatch.setattr(module, 'get_ipython', ret)
    yield ret

@pytest.fixture()
def mock_spyder_env_vars(monkeypatch):
    spyder_envs = [
        ('SPYDER_ARGS', '[]'),
        ('SPY_EXTERNAL_INTERPRETER', 'False'),
        ('SPY_UMR_ENABLED', 'True'),
        ('SPY_UMR_VERBOSE', 'True'),
        ('SPY_UMR_NAMELIST', ''),
        ('SPY_RUN_LINES_O', ''),
        ('SPY_PYLAB_O', 'True'),
        ('SPY_BACKEND_O', '0'),
        ('SPY_AUTOLOAD_PYLAB_O', 'False'),
        ('SPY_FORMAT_O', '0'),
        ('SPY_BBOX_INCHES_O', 'True'),
        ('SPY_RESOLUTION_O', '72'),
        ('SPY_WIDTH_O', '6'),
        ('SPY_HEIGHT_O', '4'),
        ('SPY_USE_FILE_O', 'False'),
        ('SPY_RUN_FILE_O', ''),
        ('SPY_AUTOCALL_O', '0'),
        ('SPY_GREEDY_O', 'False'),
        ('SPY_JEDI_O', 'False'),
        ('SPY_SYMPY_O', 'False'),
        ('SPY_TESTING', 'None'),
        ('SPY_HIDE_CMD', 'True')

    ]
    for e, v in spyder_envs:
        monkeypatch.setenv(e, v)

@pytest.fixture()
def mock_spyder_ide(mock_spyder_module, mock_get_ipython, mock_spyder_env_vars):
    yield

def test_is_spyder_ide_true(mock_spyder_ide):
    assert is_spyder_ide()

##__________________________________________________________________||
@pytest.fixture()
def mock_del_spyder_env_vars(monkeypatch):
    monkeypatch.delenv('SPYDER_ARGS', raising=False)

    envs_to_delete = [e for e in os.environ.keys() if e.startswith('SPY_')]
    for e in envs_to_delete:
        monkeypatch.delenv(e, raising=False)

def test_is_spyder_ide_false(mock_del_spyder_env_vars):
    assert not is_spyder_ide()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import atexit
import multiprocessing
import contextlib

from .machine import StateMachine

##__________________________________________________________________||
_machine = StateMachine()

##__________________________________________________________________||
def find_reporter():
    """returns the progress reporter

    This function is to be called in the main process of a
    multiprocessing program. The reporter should be registered in
    sub-processes with the function `register_reporter()`

    Returns
    -------
    object
        The progress reporter

    """
    with _machine.lock:
        _machine.state.prepare_reporter()
    return _machine.state.reporter

##__________________________________________________________________||
def register_reporter(reporter):
    """registers a reporter

    This function is to be called in sub-processes of a
    multiprocessing program.

    Parameters
    ----------
    reporter : object
        The reporter obtained in the main process by the function
        `find_reporter()`


    Returns
    -------
    None

    """
    _machine.state.register_reporter(reporter)

##__________________________________________________________________||
def flush():
    """flushes progress bars

    This function flushes all active progress bars. It returns when
    the progress bars finish updating.

    Returns
    -------
    None

    """
    with _machine.lock:
        _machine.state.flush()

##__________________________________________________________________||
def disable():
    """disables progress bars

    This function needs to be called in the main process before
    `atpbar()` or `find_reporter()` is used.

    Returns
    -------
    None

    """
    _machine.state.disable()

##__________________________________________________________________||
def end_pickup():
    """ends the pickup

    Returns
    -------
    None

    """
    with _machine.lock:
        _machine.state.end_pickup()


import multiprocessing.queues # This import prevents the issue
                              # https://github.com/alphatwirl/atpbar/issues/4

atexit.register(end_pickup)

##__________________________________________________________________||
@contextlib.contextmanager
def fetch_reporter():
    with _machine.lock:
        _machine.state.prepare_reporter()
    yield from _machine.state.fetch_reporter()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import threading
import multiprocessing

from .reporter import ProgressReporter
from .pickup import ProgressReportPickup
from .presentation.create import create_presentation
from .misc import in_main_thread
from . import detach

##__________________________________________________________________||
class StateMachine:
    def __init__(self):
        self.lock = threading.Lock()
        self.state = Initial(self)

    def change_state(self, state):
        self.state = state

##__________________________________________________________________||
class State:
    """The base class of the states
    """
    def __init__(self, machine):
        self.machine = machine

    def prepare_reporter(self):
        pass

    def register_reporter(self, reporter):
        next_state = Registered(self.machine, reporter=reporter)
        self.machine.change_state(next_state)

    def disable(self):
        self.machine.change_state(Disabled(self.machine))

    def fetch_reporter(self):
        yield None

    def flush(self):
        pass

    def end_pickup(self):
        pass

##__________________________________________________________________||
class Initial(State):
    """Initial state

    The pickup is not running.
    """

    def __init__(self, machine, reporter=None, queue=None):
        super().__init__(machine)
        self.reporter = reporter
        self.queue = queue

    def prepare_reporter(self):
        next_state = Started(self.machine, reporter=self.reporter, queue=self.queue)
        self.machine.change_state(next_state)

    def fetch_reporter(self):
        yield self.reporter

    def flush(self):
        next_state = Started(self.machine, reporter=self.reporter, queue=self.queue)
        self.machine.change_state(next_state)

class Started(State):
    """Started state

    The pickup started and is running, typically, in the main process
    """
    def __init__(self, machine, reporter=None, queue=None):
        super().__init__(machine)

        self.reporter = reporter
        self.queue = queue
        self.pickup = None
        self.pickup_owned = False

        if self.reporter is None:
            if self.queue is None:
                self.queue = multiprocessing.Queue()
            self.reporter = ProgressReporter(queue=self.queue)

        self._start_pickup()

    def _start_pickup(self):
        presentation = create_presentation()
        self.pickup = ProgressReportPickup(self.queue, presentation)
        self.pickup.start()

    def _end_pickup(self):
        self.queue.put(None)
        self.pickup.join()
        self.pickup = None
        detach.to_detach_pickup = False

    def _restart_pickup(self):
        self._end_pickup()
        self._start_pickup()

    def fetch_reporter(self):

        if not in_main_thread():
            yield self.reporter
            return

        if self.pickup_owned:
            yield self.reporter
            return

        self.pickup_owned = True

        try:
            yield self.reporter
        finally:
            with self.machine.lock:
                self.pickup_owned = False
                if detach.to_detach_pickup:
                    return
                self._restart_pickup()

    def flush(self):
        self._restart_pickup()

    def end_pickup(self):
        self._end_pickup()
        next_state = Initial(self.machine, reporter=self.reporter, queue=self.queue)
        self.machine.change_state(next_state)

##__________________________________________________________________||
class Registered(State):
    """Registered state

    Typically, in a sub-process. The reporter, which has been created
    in the main process, is registered in the sub-process
    """

    def __init__(self, machine, reporter):
        super().__init__(machine)
        self.reporter = reporter

    def fetch_reporter(self):
        yield self.reporter

class Disabled(State):
    """Disabled state
    """
    def __init__(self, machine):
        super().__init__(machine)
        self.reporter = None

##__________________________________________________________________||



import sys
import pytest

import unittest.mock as mock

from atpbar.presentation.detect.jupy import is_jupyter_notebook

##__________________________________________________________________||
@pytest.fixture()
def mock_widgets_module(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['atpbar.presentation.detect.jupy']
    monkeypatch.setattr(module, 'widgets', ret)
    yield ret

@pytest.fixture()
def mock_display_module(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['atpbar.presentation.detect.jupy']
    monkeypatch.setattr(module, 'display', ret)
    yield ret

@pytest.fixture()
def mock_get_ipython(monkeypatch):
    mock_ipython = mock.Mock()
    mock_ipython.config = {
        'IPKernelApp': {
            'connection_file': '/some/path/kernel.json',
        },
    }
    ret = mock.Mock()
    ret.return_value = mock_ipython

    module = sys.modules['atpbar.presentation.detect.jupy']
    monkeypatch.setattr(module, 'get_ipython', ret)
    yield ret

@pytest.fixture()
def mock_is_spyder_ide(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['atpbar.presentation.detect.jupy']
    monkeypatch.setattr(module, 'is_spyder_ide', ret)
    yield ret

@pytest.fixture()
def mock_jupyter_notebook(
        mock_widgets_module, mock_display_module, mock_get_ipython,
        mock_is_spyder_ide):
    mock_is_spyder_ide.return_value = False
    yield

def test_is_jupyter_notebook_true(mock_jupyter_notebook):
    assert is_jupyter_notebook()

##__________________________________________________________________||
def test_is_jupyter_notebook_false(mock_jupyter_notebook, mock_is_spyder_ide):
    mock_is_spyder_ide.return_value = True
    assert not is_jupyter_notebook()
##__________________________________________________________________||





# Tai Sakuma <tai.sakuma@gmail.com>
import ipywidgets as widgets
from IPython.display import display

from .base import Presentation

##__________________________________________________________________||
class ProgressBarJupyter(Presentation):
    def __init__(self):
        super(ProgressBarJupyter, self).__init__()
        self.interval = 0.05 # [second]

        self.container_widget = None
        self.active_box_list = [ ]
        self.complete_box_list = [ ]
        self.widget_dict = { } # {taskid: [box, bar, label]}

        self._read_time()

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def _present(self):
        self._create_widgets()
        self._update_widgets()

    def _create_widgets(self):
        if self.container_widget is None:
            self.container_widget = widgets.VBox()
            display(self.container_widget)

        for taskid in self._new_taskids:
            report = self._report_dict[taskid]
            self._create_widget(report)

    def _create_widget(self, report):
        bar = widgets.IntProgress(
            value=report['done'], min=0, max=report['total'],
            description='',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal'
         )
        label = widgets.HTML(value='')
        box = widgets.HBox([bar, label])
        self.active_box_list.append(box)
        self.container_widget.children = self.complete_box_list + self.active_box_list
        self.widget_dict[report['taskid']] = [box, bar, label]

    def _update_widgets(self):
        for taskid in self._finishing_taskids + self._active_taskids + self._new_taskids:
            report = self._report_dict[taskid]
            self._update_widget(report)

        self._reorder_widgets(report)

        if not self._new_taskids and not self._active_taskids:
            self.container_widget = None
            self.active_box_list[:] = [ ]
            self.complete_box_list[:] = [ ]
            self.widget_dict.clear()

    def _update_widget(self, report):

        percent = float(report['done'])/report['total'] if report['total'] > 0 else 1
        percent = round(percent * 100, 2)
        percent = '<pre>{:6.2f}%</pre>'.format(percent)

        box = self.widget_dict[report['taskid']][0]

        bar = self.widget_dict[report['taskid']][1]
        bar.value = report['done']
        bar.max = report['total']
        bar.description = percent
        if report['last']:
            bar.bar_style = 'success'

        label = self.widget_dict[report['taskid']][2]
        name_field_length = 32
        percent = float(report['done'])/report['total'] if report['total'] > 0 else 1
        bar = (':' * int(percent * 40)).ljust(40, " ")
        percent = round(percent * 100, 2)
        name = report['name'][0:name_field_length]
        if 'start_time' in report.keys():
            elapsed_str, remaining_str = self._get_time_track(report['start_time'], percent)
            label.value = '<pre> | {:8d} / {:8d}  ({:s} / {:s}) |: {:<{}s}</pre>'.format(report['done'], report['total'], elapsed_str, remaining_str, name, name_field_length)
        else:
            label.value = '<pre> | {:8d} / {:8d} |: {:<{}s}</pre>'.format(report['done'], report['total'], name, name_field_length)

    def _reorder_widgets(self, report):
        for taskid in self._finishing_taskids:
            box, bar, label = self.widget_dict[taskid]
            if box in self.active_box_list:
                self.active_box_list.remove(box)
                self.complete_box_list.append(box)
                self.container_widget.children = self.complete_box_list + self.active_box_list

##__________________________________________________________________||
